<?php
/**
 * @category    Payments
 * @package     Openpay_Stores
 * @author      Federico Balderas
 * @copyright   Openpay (http://openpay.mx)
 * @license     http://www.apache.org/licenses/LICENSE-2.0  Apache License Version 2.0
 */

namespace Openpay\Cards\Controller\Cards;

use Magento\Framework\App\Action\Context;
use Magento\Framework\View\Result\PageFactory;
use Openpay\Cards\Model\Payment as OpenpayPayment;
use Magento\Sales\Model\ResourceModel\Order\Invoice\Collection as InvoiceCollection;
use Magento\Sales\Model\Order\Invoice;

use Magento\Framework\App\CsrfAwareActionInterface;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\App\Request\InvalidRequestException;
use Exception;

/**
 * Webhook class
 */
class Webhook extends \Magento\Framework\App\Action\Action implements CsrfAwareActionInterface
{
    protected $request;
    protected $payment;
    protected $logger;
    protected $invoiceService;
    protected $transactionRepository;
    protected $searchCriteriaBuilder;

    public function __construct(
        Context $context,
        \Magento\Framework\App\Request\Http $request,
        OpenpayPayment $payment,
        \Openpay\Cards\Logger\Logger $logger_interface,
        \Magento\Sales\Model\Service\InvoiceService $invoiceService,
        \Magento\Sales\Api\TransactionRepositoryInterface $transactionRepository,
        \Magento\Framework\Api\SearchCriteriaBuilder $searchCriteriaBuilder
    ) {
        parent::__construct($context);
        $this->request = $request;
        $this->payment = $payment;
        $this->logger = $logger_interface;
        $this->invoiceService = $invoiceService;
        $this->transactionRepository = $transactionRepository;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
    }

    /**
     * Load the page defined in view/frontend/layout/openpay_index_webhook.xml
     * URL /openpay/index/webhook
     *
     * @return \Magento\Framework\View\Result\Page
     */
    public function execute() {
        $this->logger->debug('#webhook-cards');
        try {
            $body = file_get_contents('php://input');
            $json = json_decode($body);
            $openpay = $this->payment->getOpenpayInstance();

            $this->logger->debug("JSON_OBJECT - " . json_encode($json));
            if( (isset($json->type) && $json->type == "verification") || empty($json) || json_encode($json) == "{}"){
                header('HTTP/1.1 200 OK');
                return;
            }

            /*JSON Body Validations*/
            if(!isset($json->transaction)) throw new Exception("Transaction object not found in webhook request", 404);
            if(!isset($json->type)) throw new Exception("Charge type not found in webhook request", 404);

            /*Openpay Charge request*/
            $charge = $openpay->charges->get($json->transaction->id);
            /*Openpay Charge Validation*/
            if(!$charge) throw new Exception("Charge not found in Openpay merchant", 404);

            /*Openpay Trx Type Validation*/
            if($json->transaction->method != 'card') throw new Exception("Transaction method is not card", 500);

            /*Getting Order Data*/
            $order = $this->_objectManager->create('Magento\Sales\Model\Order');
            $order->loadByAttribute('ext_order_id', $charge->id);
            $order_status = $order->getStatus();
            $order_id = $order->getId();
            $status = \Magento\Sales\Model\Order::STATE_PROCESSING;

            /*Logging Webhook data*/
            $this->logger->debug('#webhook.data', array('webhook.trx_id' => $json->transaction->id, 'webhook.type' => $json->type , 'charge.status' => $charge->status, 'order.status' => $order_status));

            /*Magento Order validation*/
            if($order_status == 'processing' || $order_status == 'completed'){
                $this->logger->debug('#webhook.process.cancelled', array('Magento.order.status' => 'processing || completed'));
                header('HTTP/1.1 200 OK');
                exit;
            }

            /*IF transaction is only Authorization */
            if ($order && $charge->status == 'in_progress') {
                $order->setState($status)->setStatus($status);
                $order->save();
            }

            /*Update Order Status and Invoice*/
            if($json->type == 'charge.succeeded' && $charge->status == 'completed'){
                $this->logger->debug('#webhook.trx.succeeded.start', array('webhook.type' => 'charge.succeeded', 'openpay.charge.status' => 'completed' ));

                $order->setState($status)->setStatus($status);
                $order->setTotalPaid($charge->amount);
                $order->addStatusHistoryComment("Pago confirmado vía Webhook")->setIsCustomerNotified(true);
                $order->save();

                $this->searchCriteriaBuilder->addFilter('order_id', $order_id);
                $list = $this->transactionRepository->getList(
                    $this->searchCriteriaBuilder->create()
                );
                $transactions =  $list->getItems();
                foreach ($transactions as $transaction) {
                    $transaction->setIsClosed(true);
                    $transaction->save();
                }

                $requiresInvoice = true;
                /** @var InvoiceCollection $invoiceCollection */
                $invoiceCollection = $order->getInvoiceCollection();

                $this->logger->debug('#webhook.txn.id', array('$transactions', $transaction->getTxnId() ));
                $this->logger->debug('#webhook.invoiceCollection', array('$invoiceCollection', $invoiceCollection->getData() ));

                if ( $invoiceCollection->count() > 0 ) {
                    /** @var Invoice $invoice */
                    foreach ($invoiceCollection as $invoice ) {
                        $this->logger->debug('#webhook.invoice.id', array('$invoice.id', $invoice->getId(), '$invoice.incrementId' => $invoice->getIncrementId() ));
                        $this->logger->debug('#webhook.invoice.state', array('$invoice', $invoice->getState() ));
                        if ( $invoice->getState() == Invoice::STATE_OPEN) {
                            $invoice->setState(Invoice::STATE_PAID);
                            $invoice->setTransactionId($charge->id);
                            $invoice->pay()->save();
                            $requiresInvoice = false;
                            break;
                        }
                        if ($invoice->getState() == Invoice::STATE_PAID) {
                            $requiresInvoice = false;
                        }
                    }
                }

                if ( $requiresInvoice ) {
                    $invoice = $this->invoiceService->prepareInvoice($order);
                    $this->logger->debug('#webhook.requireInvoice', array('$invoice.id', $invoice->getId(), '$invoice.incrementId' => $invoice->getIncrementId() ));
                    $invoice->setTransactionId($charge->id);
                    $invoice->pay()->save();
                }
                $payment = $order->getPayment();
                $payment->setAmountPaid($charge->amount);
                $payment->setIsTransactionPending(false);
                $payment->save();

                $this->logger->debug('#webhook.trx.succeeded.end', array('Magento.order.invoice' => 'saved' ) );
            }else{
                $this->logger->debug('#webhook.trx.expired.start', array('webhook.type' => 'transaction.expired', 'openpay.charge.status' => 'expired' ));
                $order->cancel();
                $order->addStatusHistoryComment("La transacción no pudo ser procesada")->setIsCustomerNotified(true);
                $statusCanceled = $this->payment->getCustomStatus('canceled');
                $order->setStatus($statusCanceled);
                $order->save();
                $this->logger->debug('#webhook.trx.expired.end', array('Magento.order.invoice' => 'saved' ) );
            }

            header('HTTP/1.1 200 OK');

        } catch (\Exception $e) {
            $this->logger->error('#webhook-cards-exception', array('msg' => $e->getMessage(), 'code' => $e->getCode()));
            header("HTTP/1.0 500 Server Error");
        }
        exit;
    }

    /**
     * Create exception in case CSRF validation failed.
     * Return null if default exception will suffice.
     *
     * @param RequestInterface $request
     * @link https://magento.stackexchange.com/questions/253414/magento-2-3-upgrade-breaks-http-post-requests-to-custom-module-endpoint
     *
     * @return InvalidRequestException|null
     * @SuppressWarnings(PMD.UnusedFormalParameter)
     */
    public function createCsrfValidationException(RequestInterface $request): ?InvalidRequestException
    {
        return null;
    }

    /**
     * Perform custom request validation.
     * Return null if default validation is needed.
     *
     * @param RequestInterface $request
     *
     * @return bool|null
     * @SuppressWarnings(PMD.UnusedFormalParameter)
     */
    public function validateForCsrf(RequestInterface $request): ?bool
    {
        return true;
    }

}
